# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
import dataclasses
import enum
import json
import jsonschema
import time
import re

from datetime import datetime, timezone
from typing import Set, Tuple

UNIX_EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)
DTN_EPOCH = datetime(2000, 1, 1, tzinfo=timezone.utc)
UNIX_TO_DTN_OFFSET = (DTN_EPOCH - UNIX_EPOCH).total_seconds()
assert UNIX_TO_DTN_OFFSET == 946684800

# This is a quite permissive regex to check EIDs in configuration commands,
# just enforcing the general structure of <scheme>:<ssp> specified by RFC9171.
CONFIG_EID_VALIDATION_REGEX = r"^[a-zA-Z0-9]+:.+$"

# Schema for the new JSON format for configuration messages.
CONFIG_MSG_JSON_SCHEMA = {
    "type": "object",
    "properties": {
        "command": {
            "type": "string",
            "pattern": r"^(ADD|UPDATE|DELETE|QUERY)$",
        },
        "node_id": {
            "type": "string",
            "pattern": CONFIG_EID_VALIDATION_REGEX,
        },
        "cla_addr": {
            "type": "string",
            "minLength": 1,
        },
        "reachable_eids": {
            "type": "array",
            "items": {
                "type": "string",
                "pattern": CONFIG_EID_VALIDATION_REGEX,
            },
        },
        "contact_list": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "start": {"type": "integer", "minimum": 0},
                    "end": {"type": "integer", "minimum": 0},
                    "data_rate": {"type": "integer", "minimum": 0},
                    "reachable_eids": {
                        "type": "array",
                        "items": {
                            "type": "string",
                            "pattern": CONFIG_EID_VALIDATION_REGEX,
                        },
                    },
                },
                "required": ["start", "end", "data_rate"],
            }
        },
    },
    "additionalProperties": False,
    "required": ["command", "node_id"],
}


def unix2dtn(unix_timestamp):
    """Converts a given Unix timestamp into a DTN timestamp

    Its inversion is :func:`dtn2unix`.

    Args:
        unix_timestamp: Unix timestamp
    Returns:
        numeric: DTN timestamp
    """
    return unix_timestamp - UNIX_TO_DTN_OFFSET


class RouterCommand(enum.IntEnum):
    """uD3TN Command Constants"""
    ADD = 1
    UPDATE = 2
    DELETE = 3
    QUERY = 4


@dataclasses.dataclass(frozen=True)
class Contact:
    """Dataclass holding uD3TN contact information

    Attrs:
        start (int): DTN timestamp when the contact starts
        end (int): DTN timestamp when the contact is over
        bitrate (int): Bitrate of the contact, in bytes per second
        reachable_eids (Set[str]): EIDs reachable during contact; not hashed
    """
    start: int
    end: int
    bitrate: int
    # The following fields are not part of the hash when stored in a set.
    reachable_eids: Set[str] = dataclasses.field(
        compare=False,
        default_factory=set,
    )

    def __repr__(self):
        return (
            f"<Contact start={self.start}, end={self.end}, "
            f"bitrate={self.bitrate}>"
        )

    def __str__(self):
        if self.reachable_eids:
            eid_list = "[" + ",".join(
                "(" + eid + ")" for eid in self.reachable_eids
            ) + "]"
            return f"{{{self.start},{self.end},{self.bitrate},{eid_list}}}"
        else:
            return f"{{{self.start},{self.end},{self.bitrate}}}"


def make_contact(start_offset, duration, bitrate, reachable_eids=None):
    """Create a :class:`Contact` tuple relative to the current time

    NOTE: The parameters of this function are also compatible with float-typed
          input arguments.

    Args:
        start_offset (int): Start point of the contact in seconds from now
        duration (int): Duration of the contact in seconds
        bitrate (int): Bitrate of the contact, in bytes per second
        reachable_eids (Set[str]): EIDs reachable during contact; not hashed
    Returns:
        Contact: contact tuple with DTN timestamps
    """
    cur_time = time.time()
    start = unix2dtn(cur_time + start_offset)

    if reachable_eids:
        return Contact(
            start=int(round(start)),
            end=int(round(start + duration)),
            bitrate=int(round(bitrate)),
            reachable_eids=set(reachable_eids),
        )
    else:
        return Contact(
            start=int(round(start)),
            end=int(round(start + duration)),
            bitrate=int(round(bitrate)),
        )


class ConfigMessage(object):
    """Base class for the uD3TN configuration messages, which can be processed
    by the config endpoint of its deterministic first-contact forwarding BDM
    (either integrated or external).

    Args:
        eid (Optional[str]): The endpoint identifier of a contact
        cla_address (Optional[str]): The Convergency Layer Adapter (CLA)
            address for the contact's EID
        reachable_eids (Iterable[str], optional): List of reachable EIDs via
            this contact
        contacts (Iterable[Contact], optional): List of contacts with the node
        type (RouterCommand, optional): Type of the configuration message (add,
            remove, ...)
    """

    def __init__(self, eid, cla_address, reachable_eids=None, contacts=None,
                 type=RouterCommand.ADD):
        self.eid = eid
        self.cla_address = cla_address
        self.reachable_eids = set(reachable_eids) if reachable_eids else set()
        self.contacts = set(contacts) if contacts else set()
        self.type = type

    def __repr__(self):
        return "<ConfigMessage {!r} {!r} reachable={} contacts={}>".format(
            self.eid, self.cla_address, self.reachable_eids, self.contacts
        )

    @staticmethod
    def parse(config_str: str, schema_validate: bool = True
              ) -> "LegacyConfigMessage":
        """Parse the provided configuration string representation,
        automatically decising whether it is a JSON or legacy representation.

        Args:
            config_str (str): The serialized string representation, obtained
                e.g. via str() applied on a LegacyConfigMessage or
                JSONConfigMessage object.

        Return:
            A ConfigMessage instance of the correct subclass representing the
            data parsed from config_str.
        """

        if config_str[0] == "{":
            return JSONConfigMessage.parse(config_str, schema_validate)
        else:
            return LegacyConfigMessage.parse(config_str)


class JSONConfigMessage(ConfigMessage):
    """uD3TN configuration message using the JSON encoding.

    NOTE: At the moment (v0.15.0), for this to be supported, the external
        (Python) forwarding module has to be used.
    """

    def __str__(self):
        ret_obj = {
            "command": {
                1: "ADD",
                2: "UPDATE",
                3: "DELETE",
                4: "QUERY",
            }[self.type],
            "node_id": self.eid,
        }

        if self.cla_address:
            ret_obj["cla_addr"] = self.cla_address

        if self.reachable_eids:
            ret_obj["reachable_eids"] = list(self.reachable_eids)

        if self.contacts:
            ret_obj["contact_list"] = [
                {
                    "start": c.start,
                    "end": c.end,
                    "data_rate": c.bitrate,
                    "reachable_eids": (
                        list(c.reachable_eids) if c.reachable_eids else []
                    ),
                }
                for c in self.contacts
            ]

        return json.dumps(ret_obj)

    def __bytes__(self):
        return str(self).encode('utf-8')

    def to_legacy_format(self) -> "LegacyConfigMessage":
        return LegacyConfigMessage(
            self.eid,
            self.cla_address,
            self.reachable_eids,
            self.contacts,
            type=self.type,
        )

    @staticmethod
    def parse(config_str: str, schema_validate: bool = True
              ) -> "JSONConfigMessage":
        """Parse the provided JSON configuration string representation.

        Args:
            config_str (str): The serialized string representation, obtained
                e.g. via str() applied on a JSONConfigMessage object.

        Return:
            A JSONConfigMessage instance representing the data parsed from
            config_str.
        """

        obj = json.loads(config_str)
        if schema_validate:
            jsonschema.validate(obj, CONFIG_MSG_JSON_SCHEMA)

        return JSONConfigMessage(
            obj["node_id"],
            obj["cla_addr"],
            obj.get("reachable_eids", []),
            [
                Contact(
                    start=c["start"],
                    end=c["end"],
                    bitrate=c["data_rate"],
                    reachable_eids=c.get("reachable_eids", None),
                )
                for c in obj.get("contact_list", [])
            ],
            type=RouterCommand({
                "ADD": 1,
                "UPDATE": 2,
                "DELETE": 3,
                "QUERY": 4,
            }[obj["command"]]),
        )


class LegacyConfigMessage(ConfigMessage):
    """uD3TN configuration message using the legacy custom format."""

    def __str__(self):
        # missing escaping has to be addressed in uD3TN
        for part in ({str(self.eid), str(self.cla_address)} |
                     self.reachable_eids):
            assert "(" not in part, "unsupported character in string"
            assert ")" not in part, "unsupported character in string"

        assert "," not in str(self.eid), "unsupported character in string"

        if self.reachable_eids:
            eid_list = "[" + ",".join(
                "(" + eid + ")" for eid in self.reachable_eids
            ) + "]"
        else:
            eid_list = ""

        if self.contacts:
            contact_list = (
                    "[" +
                    ",".join(
                        str(contact)
                        for contact in self.contacts
                    ) +
                    "]"
                )
        else:
            contact_list = ""

        if self.eid is None:
            return "{};".format(self.type)
        elif self.cla_address is None:
            return "{}({})::{}:{};".format(
                self.type,
                self.eid,
                eid_list,
                contact_list,
            )
        else:
            return "{}({}):({}):{}:{};".format(
                self.type,
                self.eid,
                self.cla_address,
                eid_list,
                contact_list,
            )

    def __bytes__(self):
        return str(self).encode('ascii')

    def to_json_format(self) -> JSONConfigMessage:
        return JSONConfigMessage(
            self.eid,
            self.cla_address,
            self.reachable_eids,
            self.contacts,
            type=self.type,
        )

    @staticmethod
    def parse(config_str: str):
        """Parse the provided configuration string representation.

        Args:
            config_str (str): The serialized string representation, obtained
                e.g. via str() applied on a LegacyConfigMessage object.

        Return:
            A LegacyConfigMessage instance representing the data parsed from
            config_str.
        """
        cmd_type, node_id, cla_addr, reachable_eids, contacts = _parse_config(
            config_str,
        )
        return LegacyConfigMessage(
            node_id,
            cla_addr,
            reachable_eids,
            contacts,
            type=RouterCommand(cmd_type),
        )


# Config Parser helper functions
CONFIG_PARSE_REGEX = re.compile(
    r"^(?P<cmd>[1-4])(\((?P<nodeid>[^\)]+)\))?(\:(\((?P<claaddr>[^\)]+)\))?"
    r"(\:(?P<eidlist>\[(\(([^\)]+)\),?)*\])?"
    r"(\:(?P<contactlist>\[(\{[0-9]+,[0-9]+,[0-9]+(,\[(\([^\)]+\),?)+\])?\}"
    r",?)*\])?)?)?)?;$"
)
EID_LIST_PARSE_REGEX = re.compile(
    r"^[\[,]\((?P<eid>[^\)]+)\)(?P<remainder>,\(.+\))?\]?$"
)
CONTACT_LIST_PARSE_REGEX = re.compile(
    r"^[\[,]\{(?P<start>[0-9]+),(?P<end>[0-9]+),(?P<datarate>[0-9]+)"
    r"(,(?P<eidlist>\[(\([^\)]+\),?)+\]))?\}(?P<remainder>,\{.+\})?\]?$"
)


def _parse_eid_list(eid_list: str) -> Set[str]:
    if eid_list is None:
        return set()
    match_obj = EID_LIST_PARSE_REGEX.match(eid_list)
    if match_obj is None:
        return set()
    if match_obj.group(2) is None:
        return {match_obj.group(1)}
    return {match_obj.group(1)} | _parse_eid_list(match_obj.group(2))


def _parse_contact_list(contact_list: str) -> Set[Contact]:
    if contact_list is None:
        return set()
    match_obj = CONTACT_LIST_PARSE_REGEX.match(contact_list)
    if match_obj is None:
        return set()
    contact = Contact(
        start=int(match_obj.group("start")),
        end=int(match_obj.group("end")),
        bitrate=int(match_obj.group("datarate")),
        reachable_eids=_parse_eid_list(match_obj.group("eidlist")),
    )
    if match_obj.group("remainder") is None:
        return {contact}
    return {contact} | _parse_contact_list(
        match_obj.group("remainder"),
    )


def _parse_config(config_str: str) -> Tuple[int, str, Set[str], Set[Contact]]:
    match_obj = CONFIG_PARSE_REGEX.match(config_str)
    if match_obj is None:
        raise ValueError("invalid config string")
    cmd_type = RouterCommand(int(match_obj.group("cmd")))
    node_id = match_obj.group("nodeid")
    if node_id is None and cmd_type != RouterCommand.QUERY:
        raise ValueError("an empty node ID is only valid for QUERY commands")
    return (
        cmd_type,
        node_id,
        match_obj.group("claaddr"),
        _parse_eid_list(match_obj.group("eidlist")),
        _parse_contact_list(match_obj.group("contactlist")),
    )


def test_parse_config():
    assert _parse_config(
        "1(dtn://ud3tn2.dtn/):(mtcp:127.0.0.1:4223)::["
        "{1401519306972,1401519316972,1200,[(dtn://89326/),(dtn://12349/)]},"
        "{1401519506972,1401519516972,1200,[(dtn://89326/),(dtn://12349/)]}];"
    ) == (
        1,
        "dtn://ud3tn2.dtn/",
        "mtcp:127.0.0.1:4223",
        set(),
        {
            Contact(
                start=1401519306972,
                end=1401519316972,
                bitrate=1200,
                reachable_eids={
                    "dtn://89326/",
                    "dtn://12349/",
                },
            ),
            Contact(
                start=1401519506972,
                end=1401519516972,
                bitrate=1200,
                reachable_eids={
                    "dtn://89326/",
                    "dtn://12349/",
                },
            ),
        },
    )
    assert _parse_config(
        "1(dtn://ud3tn2.dtn/)::[(dtn://18471/),(dtn://81491/)]:"
        "[{1401519406972,1401819306972,1200}];"
    ) == (
        1,
        "dtn://ud3tn2.dtn/",
        None,
        {"dtn://18471/", "dtn://81491/"},
        {
            Contact(
                start=1401519406972,
                end=1401819306972,
                bitrate=1200,
                reachable_eids=set(),
            ),
        },
    )
    assert _parse_config(
        "2(dtn://ud3tn2.dtn/):(mtcp:127.0.0.1:4223):"
        "[(dtn://89326/),(dtn://12349/)];"
    ) == (
        2,
        "dtn://ud3tn2.dtn/",
        "mtcp:127.0.0.1:4223",
        {"dtn://89326/", "dtn://12349/"},
        set(),
    )
    assert _parse_config(
        "2(dtn://ud3tn2.dtn/)::[]:[];"
    ) == (
        2,
        "dtn://ud3tn2.dtn/",
        None,
        set(),
        set(),
    )
    assert _parse_config(
        "2(ipn:1.0)::[];"
    ) == (
        2,
        "ipn:1.0",
        None,
        set(),
        set(),
    )
    assert _parse_config(
        "3(dtn://ud3tn2.dtn/);"
    ) == (
        3,
        "dtn://ud3tn2.dtn/",
        None,
        set(),
        set(),
    )
    assert _parse_config(
        "1(dtn://13714/):(tcpspp:):[(dtn://18471/),(dtn://81491/)]:;"
    ) == (
        1,
        "dtn://13714/",
        "tcpspp:",
        {"dtn://18471/", "dtn://81491/"},
        set(),
    )
    assert _parse_config(
        "4(dtn://ud3tn2.dtn/);"
    ) == (
        4,
        "dtn://ud3tn2.dtn/",
        None,
        set(),
        set(),
    )
    assert _parse_config(
        "4;"
    ) == (
        4,
        None,
        None,
        set(),
        set(),
    )


def test_parse_and_serialize():
    config_str1 = (
        "1(dtn://ud3tn2.dtn/):(mtcp:127.0.0.1:4223):[(ipn:1.0)]:["
        "{1401519306972,1401519316972,2400,[(dtn://89326/),(dtn://66553/)]},"
        "{1401519506972,1401519516972,1200,[(dtn://89326/),(dtn://12349/)]}];"
    )
    cm1 = ConfigMessage.parse(config_str1)
    assert isinstance(cm1, LegacyConfigMessage)
    # NOTE: We cannot `assert config_str1 == str(cm1)` here as the set
    # ordering may differ.
    cm2 = ConfigMessage.parse(str(cm1))
    assert isinstance(cm2, LegacyConfigMessage)
    assert cm1.type == cm2.type == 1
    assert cm1.eid == cm2.eid == "dtn://ud3tn2.dtn/"
    assert cm1.cla_address == cm2.cla_address == "mtcp:127.0.0.1:4223"
    assert cm1.reachable_eids == cm2.reachable_eids == {"ipn:1.0"}
    assert cm1.contacts == cm2.contacts
    # Test JSON conversion and serialization
    jo = cm1.to_json_format()
    assert isinstance(jo, JSONConfigMessage)
    jstr = str(jo)
    jo2 = ConfigMessage.parse(jstr, True)
    assert cm1.type == jo2.type == 1
    assert cm1.eid == jo2.eid == "dtn://ud3tn2.dtn/"
    assert cm1.cla_address == jo2.cla_address == "mtcp:127.0.0.1:4223"
    assert cm1.reachable_eids == jo2.reachable_eids == {"ipn:1.0"}
    assert cm1.contacts == jo2.contacts
    cm3 = jo2.to_legacy_format()
    assert isinstance(cm3, LegacyConfigMessage)
    assert cm1.type == cm3.type == 1
    assert cm1.eid == cm3.eid == "dtn://ud3tn2.dtn/"
    assert cm1.cla_address == cm3.cla_address == "mtcp:127.0.0.1:4223"
    assert cm1.reachable_eids == cm3.reachable_eids == {"ipn:1.0"}
    assert cm1.contacts == cm3.contacts



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: storage_agent.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'storage_agent.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13storage_agent.proto\"\x8c\x01\n\x10\x43ompoundBundleId\x12\x12\n\nsource_eid\x18\x01 \x01(\t\x12\x1a\n\x12\x63reation_timestamp\x18\x03 \x01(\x04\x12\x17\n\x0fsequence_number\x18\x04 \x01(\x04\x12\x17\n\x0f\x66ragment_offset\x18\x05 \x01(\x04\x12\x16\n\x0epayload_length\x18\x06 \x01(\x04\":\n\x14\x42undleMetadataFilter\x12\x15\n\x08\x65id_glob\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x0b\n\t_eid_glob\"\x90\x01\n\x0bStorageCall\x12$\n\toperation\x18\x01 \x01(\x0e\x32\x11.StorageOperation\x12\x1f\n\x02id\x18\x02 \x01(\x0b\x32\x11.CompoundBundleIdH\x00\x12)\n\x08metadata\x18\x04 \x01(\x0b\x32\x15.BundleMetadataFilterH\x00\x42\x0f\n\rbundle_filter*\xa3\x01\n\x10StorageOperation\x12!\n\x1dSTORAGE_OPERATION_UNSPECIFIED\x10\x00\x12\"\n\x1eSTORAGE_OPERATION_LIST_BUNDLES\x10\x01\x12$\n STORAGE_OPERATION_DELETE_BUNDLES\x10\x02\x12\"\n\x1eSTORAGE_OPERATION_PUSH_BUNDLES\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'storage_agent_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STORAGEOPERATION']._serialized_start=374
  _globals['_STORAGEOPERATION']._serialized_end=537
  _globals['_COMPOUNDBUNDLEID']._serialized_start=24
  _globals['_COMPOUNDBUNDLEID']._serialized_end=164
  _globals['_BUNDLEMETADATAFILTER']._serialized_start=166
  _globals['_BUNDLEMETADATAFILTER']._serialized_end=224
  _globals['_STORAGECALL']._serialized_start=227
  _globals['_STORAGECALL']._serialized_end=371
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging
import sys

import cbor2  # type: ignore

from pyd3tn.bundle7 import Bundle
from ud3tn_utils.aap2 import (
    AAP2UnixClient,
    AAP2TCPClient,
    AAP2ServerDisconnected,
    BundleADUFlags,
    ResponseStatus,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    add_keepalive_parser_argument,
    get_secret_from_args,
    initialize_logger,
)


logger = logging.getLogger(__name__)


def run_aap_recv(aap2_client, max_count, output, verify_pl, newline):
    logger.info("Waiting for bundles...")
    counter = 0

    while True:
        try:
            msg = aap2_client.receive_msg()
        except AAP2ServerDisconnected:
            logger.warning("µD3TN has closed the connection.")
            sys.exit(1)
        except KeyboardInterrupt:
            logger.info("Terminated by keyboard interrupt.")
            sys.exit(130)  # exit status for SIGINT

        msg_type = msg.WhichOneof("msg")
        if msg_type == "keepalive":
            logger.debug("Received keepalive message, acknowledging.")
            aap2_client.send_response_status(
                ResponseStatus.RESPONSE_STATUS_ACK
            )
            continue
        elif msg_type != "adu":
            logger.info("Received message with field '%s' set, discarding.",
                        msg_type)
            continue

        adu_msg, bundle_data = aap2_client.receive_adu(msg.adu)
        aap2_client.send_response_status(
            ResponseStatus.RESPONSE_STATUS_SUCCESS
        )

        enc = False
        err = False
        sr = False

        if BundleADUFlags.BUNDLE_ADU_BPDU in adu_msg.adu_flags:
            payload = cbor2.loads(bundle_data)
            bundle = Bundle.parse(payload[2])
            payload = bundle.payload_block.data
            enc = True
        else:
            payload = bundle_data
            sr = BundleADUFlags.BUNDLE_ADU_STATUS_REPORT in adu_msg.adu_flags

        if not err:
            logger.info(
                "Received%s %s from '%s', payload len = %d",
                " encapsulated" if enc else "",
                "Status Report" if sr else "Bundle",
                msg.adu.src_eid,
                len(payload),
            )
            output.write(payload)
            if newline:
                output.write(b"\n")
            output.flush()
            if verify_pl is not None and verify_pl.encode("utf-8") != payload:
                logger.fatal("Unexpected payload != '%s'", verify_pl)
                sys.exit(1)
        else:
            logger.warning(
                "Received administrative record of unknown type from '%s'!",
                msg.adu.src_eid
            )

        counter += 1
        if max_count and counter >= max_count:
            logger.info("Expected amount of bundles received, terminating.")
            return


def main():
    parser = argparse.ArgumentParser(
        description="register an agent with uD3TN and wait for bundles",
    )

    add_common_parser_arguments(parser)
    add_keepalive_parser_argument(parser)

    parser.add_argument(
        "-c", "--count",
        type=int,
        default=None,
        help="amount of bundles to be received before terminating",
    )
    parser.add_argument(
        "-o", "--output",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
        help="file to write the received bundle contents",
    )
    parser.add_argument(
        "--verify-pl",
        default=None,
        help="verify that the payload is equal to the provided string",
    )
    parser.add_argument(
        "--newline",
        action="store_true",
        help="print a line feed character after every received bundle payload",
    )

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    try:
        if args.tcp:
            aap2_client = AAP2TCPClient(address=args.tcp)
        else:
            aap2_client = AAP2UnixClient(address=args.socket)

        with aap2_client:
            secret = aap2_client.configure(
                args.agentid,
                subscribe=True,
                secret=get_secret_from_args(args),
                keepalive_seconds=args.keepalive_seconds,
            )
            logger.info("Assigned agent secret: '%s'", secret)
            run_aap_recv(
                aap2_client,
                args.count,
                args.output,
                args.verify_pl,
                args.newline,
            )
    finally:
        args.output.close()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: aap2.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'aap2.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\naap2.proto\x12\x04\x61\x61p2\"\xf0\x01\n\nAAPMessage\x12 \n\x07welcome\x18\x01 \x01(\x0b\x32\r.aap2.WelcomeH\x00\x12(\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x16.aap2.ConnectionConfigH\x00\x12\x1e\n\x03\x61\x64u\x18\x03 \x01(\x0b\x32\x0f.aap2.BundleADUH\x00\x12-\n\x0e\x64ispatch_event\x18\x04 \x01(\x0b\x32\x13.aap2.DispatchEventH\x00\x12\x1a\n\x04link\x18\x05 \x01(\x0b\x32\n.aap2.LinkH\x00\x12$\n\tkeepalive\x18\x06 \x01(\x0b\x32\x0f.aap2.KeepaliveH\x00\x42\x05\n\x03msg\"\x1a\n\x07Welcome\x12\x0f\n\x07node_id\x18\x01 \x01(\t\"\x8c\x01\n\x10\x43onnectionConfig\x12\x15\n\ris_subscriber\x18\x01 \x01(\x08\x12!\n\tauth_type\x18\x02 \x01(\x0e\x32\x0e.aap2.AuthType\x12\x0e\n\x06secret\x18\x03 \x01(\t\x12\x13\n\x0b\x65ndpoint_id\x18\x04 \x01(\t\x12\x19\n\x11keepalive_seconds\x18\x05 \x01(\r\"\xbd\x01\n\tBundleADU\x12\x0f\n\x07src_eid\x18\x01 \x01(\t\x12\x0f\n\x07\x64st_eid\x18\x02 \x01(\t\x12\x15\n\rreport_to_eid\x18\x07 \x01(\t\x12\x1d\n\x15\x63reation_timestamp_ms\x18\x03 \x01(\x04\x12\x17\n\x0fsequence_number\x18\x04 \x01(\x04\x12\x16\n\x0epayload_length\x18\x05 \x01(\x04\x12\'\n\tadu_flags\x18\x06 \x03(\x0e\x32\x14.aap2.BundleADUFlags\"\xc2\x01\n\x06\x42undle\x12\x0f\n\x07src_eid\x18\x01 \x01(\t\x12\x0f\n\x07\x64st_eid\x18\x02 \x01(\t\x12\x1d\n\x15\x63reation_timestamp_ms\x18\x03 \x01(\x04\x12\x17\n\x0fsequence_number\x18\x04 \x01(\x04\x12\x16\n\x0epayload_length\x18\x05 \x01(\x04\x12\x17\n\x0f\x66ragment_offset\x18\x06 \x01(\x04\x12\x18\n\x10total_adu_length\x18\x07 \x01(\x04\x12\x13\n\x0blifetime_ms\x18\x08 \x01(\x04\"\xfa\x01\n\x12\x42undleDispatchInfo\x12\x17\n\x0fserialized_size\x18\x01 \x01(\x04\x12\x1b\n\x13min_frag_size_first\x18\x02 \x01(\x04\x12\x1a\n\x12min_frag_size_last\x18\x03 \x01(\x04\x12\x18\n\x10\x64ispatch_time_ms\x18\x04 \x01(\x04\x12\x1a\n\x12\x65xpiration_time_ms\x18\x05 \x01(\x04\x12\x1d\n\x15\x64ispatched_to_node_id\x18\x06 \x01(\t\x12\x1e\n\x16\x64ispatched_to_cla_addr\x18\x07 \x01(\t\x12\x1d\n\x15max_bundle_size_bytes\x18\x08 \x01(\x04\"\x8d\x01\n\rDispatchEvent\x12\x1c\n\x06\x62undle\x18\x01 \x01(\x0b\x32\x0c.aap2.Bundle\x12$\n\x06reason\x18\x02 \x01(\x0e\x32\x14.aap2.DispatchReason\x12\x38\n\x16\x61\x64\x64itional_information\x18\x03 \x01(\x0b\x32\x18.aap2.BundleDispatchInfo\"t\n\x04Link\x12 \n\x06status\x18\x01 \x01(\x0e\x32\x10.aap2.LinkStatus\x12\x14\n\x0cpeer_node_id\x18\x02 \x01(\t\x12\x15\n\rpeer_cla_addr\x18\x03 \x01(\t\x12\x1d\n\x04\x66lag\x18\x04 \x01(\x0e\x32\x0f.aap2.LinkFlags\"\x0b\n\tKeepalive\"\x91\x01\n\x0b\x41\x41PResponse\x12-\n\x0fresponse_status\x18\x01 \x01(\x0e\x32\x14.aap2.ResponseStatus\x12-\n\x0f\x64ispatch_result\x18\x02 \x01(\x0b\x32\x14.aap2.DispatchResult\x12$\n\x0e\x62undle_headers\x18\x03 \x01(\x0b\x32\x0c.aap2.Bundle\"\x99\x01\n\x0e\x44ispatchResult\x12\x34\n\tnext_hops\x18\x01 \x03(\x0b\x32!.aap2.DispatchResult.NextHopEntry\x1aQ\n\x0cNextHopEntry\x12\x0f\n\x07node_id\x18\x01 \x01(\t\x12\x17\n\x0f\x66ragment_offset\x18\x02 \x01(\x04\x12\x17\n\x0f\x66ragment_length\x18\x03 \x01(\x04*{\n\x08\x41uthType\x12\x15\n\x11\x41UTH_TYPE_DEFAULT\x10\x00\x12\x19\n\x15\x41UTH_TYPE_FIB_CONTROL\x10\x01\x12\x1d\n\x19\x41UTH_TYPE_BUNDLE_DISPATCH\x10\x02\x12\x1e\n\x1a\x41UTH_TYPE_FIB_AND_DISPATCH\x10\x03*x\n\x0e\x42undleADUFlags\x12\x15\n\x11\x42UNDLE_ADU_NORMAL\x10\x00\x12\x13\n\x0f\x42UNDLE_ADU_BPDU\x10\x01\x12\x1c\n\x18\x42UNDLE_ADU_WITH_BDM_AUTH\x10\x02\x12\x1c\n\x18\x42UNDLE_ADU_STATUS_REPORT\x10\x03*\xdf\x01\n\x0e\x44ispatchReason\x12\x1f\n\x1b\x44ISPATCH_REASON_UNSPECIFIED\x10\x00\x12 \n\x1c\x44ISPATCH_REASON_NO_FIB_ENTRY\x10\x01\x12!\n\x1d\x44ISPATCH_REASON_LINK_INACTIVE\x10\x02\x12&\n\"DISPATCH_REASON_CLA_ENQUEUE_FAILED\x10\x03\x12\x1d\n\x19\x44ISPATCH_REASON_TX_FAILED\x10\x04\x12 \n\x1c\x44ISPATCH_REASON_TX_SUCCEEDED\x10\x05*u\n\nLinkStatus\x12\x1b\n\x17LINK_STATUS_UNSPECIFIED\x10\x00\x12\x12\n\x0eLINK_STATUS_UP\x10\x03\x12\x14\n\x10LINK_STATUS_DOWN\x10\x04\x12 \n\x1cLINK_STATUS_DROP_CLA_MAPPING\x10\x05*5\n\tLinkFlags\x12\x12\n\x0eLINK_FLAG_NONE\x10\x00\x12\x14\n\x10LINK_FLAG_DIRECT\x10\x01*\x85\x02\n\x0eResponseStatus\x12\x1f\n\x1bRESPONSE_STATUS_UNSPECIFIED\x10\x00\x12\x1b\n\x17RESPONSE_STATUS_SUCCESS\x10\x01\x12\x17\n\x13RESPONSE_STATUS_ACK\x10\x02\x12\x19\n\x15RESPONSE_STATUS_ERROR\x10\x08\x12\x1b\n\x17RESPONSE_STATUS_TIMEOUT\x10\t\x12#\n\x1fRESPONSE_STATUS_INVALID_REQUEST\x10\n\x12\x1d\n\x19RESPONSE_STATUS_NOT_FOUND\x10\x0b\x12 \n\x1cRESPONSE_STATUS_UNAUTHORIZED\x10\x0c\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'aap2_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_AUTHTYPE']._serialized_start=1655
  _globals['_AUTHTYPE']._serialized_end=1778
  _globals['_BUNDLEADUFLAGS']._serialized_start=1780
  _globals['_BUNDLEADUFLAGS']._serialized_end=1900
  _globals['_DISPATCHREASON']._serialized_start=1903
  _globals['_DISPATCHREASON']._serialized_end=2126
  _globals['_LINKSTATUS']._serialized_start=2128
  _globals['_LINKSTATUS']._serialized_end=2245
  _globals['_LINKFLAGS']._serialized_start=2247
  _globals['_LINKFLAGS']._serialized_end=2300
  _globals['_RESPONSESTATUS']._serialized_start=2303
  _globals['_RESPONSESTATUS']._serialized_end=2564
  _globals['_AAPMESSAGE']._serialized_start=21
  _globals['_AAPMESSAGE']._serialized_end=261
  _globals['_WELCOME']._serialized_start=263
  _globals['_WELCOME']._serialized_end=289
  _globals['_CONNECTIONCONFIG']._serialized_start=292
  _globals['_CONNECTIONCONFIG']._serialized_end=432
  _globals['_BUNDLEADU']._serialized_start=435
  _globals['_BUNDLEADU']._serialized_end=624
  _globals['_BUNDLE']._serialized_start=627
  _globals['_BUNDLE']._serialized_end=821
  _globals['_BUNDLEDISPATCHINFO']._serialized_start=824
  _globals['_BUNDLEDISPATCHINFO']._serialized_end=1074
  _globals['_DISPATCHEVENT']._serialized_start=1077
  _globals['_DISPATCHEVENT']._serialized_end=1218
  _globals['_LINK']._serialized_start=1220
  _globals['_LINK']._serialized_end=1336
  _globals['_KEEPALIVE']._serialized_start=1338
  _globals['_KEEPALIVE']._serialized_end=1349
  _globals['_AAPRESPONSE']._serialized_start=1352
  _globals['_AAPRESPONSE']._serialized_end=1497
  _globals['_DISPATCHRESULT']._serialized_start=1500
  _globals['_DISPATCHRESULT']._serialized_end=1653
  _globals['_DISPATCHRESULT_NEXTHOPENTRY']._serialized_start=1572
  _globals['_DISPATCHRESULT_NEXTHOPENTRY']._serialized_end=1653
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging
import sys

from google.protobuf.internal import encoder

from ud3tn_utils.aap2 import (
    AAP2TCPClient,
    AAP2UnixClient,
    AuthType,
    BundleADU,
    BundleADUFlags,
    ResponseStatus,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    argparse_non_empty_str,
    get_secret_from_args,
    initialize_logger,
)
from ud3tn_utils.storage_agent import StorageCall, StorageOperation


logger = logging.getLogger(__name__)


def _init_parser():
    parser = argparse.ArgumentParser(
        description="send StorageAgent commands via uD3TN's AAP2 interface",
    )

    add_common_parser_arguments(parser)

    parser.add_argument(
        "--storage-agent-eid",
        type=argparse_non_empty_str,
        default="dtn://ud3tn.dtn/sqlite",
        help="destination EID (dtn or ipn URI) of the StorageAgent",
    )

    subparser = parser.add_subparsers(dest="cmd")

    delete_parser = subparser.add_parser("delete")
    delete_parser.add_argument(
        "--dest-eid-glob",
        type=argparse_non_empty_str,
        help="delete bundles that match the specified destination EID glob"
    )
    delete_parser.add_argument(
        "--compound-bundle-id",
        type=str,
        nargs=6,
        metavar=(
            'SOURCE',
            'DESTINATION',
            'CREATION_TIMESTAMP',
            'SEQUENCE_NUMBER',
            'FRAGMENT_OFFSET',
            'PAYLOAD_LENGTH'
        ),
        help="delete the bundle that matches the compound bundle id"
    )

    push_parser = subparser.add_parser("push")
    push_parser.add_argument(
        "--dest-eid-glob",
        type=argparse_non_empty_str,
        help="push bundles that match the specified destination EID glob"
    )
    push_parser.add_argument(
        "--compound-bundle-id",
        type=str,
        nargs=6,
        metavar=(
            'SOURCE',
            'DESTINATION',
            'CREATION_TIMESTAMP',
            'SEQUENCE_NUMBER',
            'FRAGMENT_OFFSET',
            'PAYLOAD_LENGTH'
        ),
        help="push the bundle that match the compound bundle id"
    )

    return parser


def main():
    parser = _init_parser()
    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    cmd = StorageCall()
    if args.cmd == "delete":
        cmd.operation = StorageOperation.STORAGE_OPERATION_DELETE_BUNDLES
    elif args.cmd == "push":
        cmd.operation = StorageOperation.STORAGE_OPERATION_PUSH_BUNDLES
    else:
        logger.error(f"Invalid `cmd` argument: {args.cmd}")
        parser.print_help()
        sys.exit(1)

    if args.compound_bundle_id:
        cmd.id.source_eid = args.compound_bundle_id[0]
        cmd.id.destination_eid = args.compound_bundle_id[1]
        cmd.id.creation_timestamp = int(args.compound_bundle_id[2])
        cmd.id.sequence_number = int(args.compound_bundle_id[3])
        cmd.id.fragment_offset = int(args.compound_bundle_id[4])
        cmd.id.payload_length = int(args.compound_bundle_id[5])
    elif args.dest_eid_glob:
        cmd.metadata.eid_glob = args.dest_eid_glob
    else:
        logger.error("Missing filter")
        parser.print_help()
        sys.exit(1)

    cmd_bytes = cmd.SerializeToString()
    proto_msg = encoder._VarintBytes(len(cmd_bytes)) + cmd_bytes

    if args.tcp:
        aap2_client = AAP2TCPClient(address=args.tcp)
    else:
        aap2_client = AAP2UnixClient(address=args.socket)
    with aap2_client:
        secret = aap2_client.configure(
            args.agentid,
            subscribe=False,
            secret=get_secret_from_args(args),
            auth_type=AuthType.AUTH_TYPE_BUNDLE_DISPATCH,
        )

        logger.info("Assigned agent secret: '%s'", secret)

        aap2_client.send_adu(
            BundleADU(
                dst_eid=args.storage_agent_eid,
                payload_length=len(proto_msg),
                adu_flags=[BundleADUFlags.BUNDLE_ADU_WITH_BDM_AUTH],
            ),
            proto_msg,
        )

        response = aap2_client.receive_response()
        if response.response_status != ResponseStatus.RESPONSE_STATUS_SUCCESS:
            logger.warning(
                "µD3TN responded with unexpected status: %s (%d)",
                aap2_client.response_status_name(response.response_status),
                response.response_status,
            )
            sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging
import sys

import cbor2  # type: ignore

from pyd3tn.bundle7 import Bundle
from ud3tn_utils.aap import AAPUnixClient, AAPTCPClient
from ud3tn_utils.aap.aap_message import AAPMessageType
from ud3tn_utils.aap.bin.helpers import (
    add_common_parser_arguments,
    initialize_logger,
)


logger = logging.getLogger(__name__)


def run_aap_recv(aap_client, max_count, output, verify_pl, newline):
    logger.info("Waiting for bundles...")
    counter = 0

    while True:
        try:
            msg = aap_client.receive()
        except KeyboardInterrupt:
            logger.info("Terminated by keyboard interrupt.")
            sys.exit(130)  # exit status for SIGINT
        if not msg:
            logger.warning("µD3TN has closed the connection.")
            sys.exit(1)

        enc = False
        err = False
        if msg.msg_type == AAPMessageType.RECVBUNDLE:
            payload = msg.payload
        elif msg.msg_type == AAPMessageType.RECVBIBE:
            payload = cbor2.loads(msg.payload)
            bundle = Bundle.parse(payload[2])
            payload = bundle.payload_block.data
            enc = True

        if not err:
            enc = " encapsulated" if enc else ""
            logger.info(
                "Received%s bundle from '%s', payload len = %d",
                enc,
                msg.eid,
                len(payload),
            )
            output.write(payload)
            if newline:
                output.write(b"\n")
            output.flush()
            if verify_pl is not None and verify_pl.encode("utf-8") != payload:
                logger.fatal("Unexpected payload != '%s'", verify_pl)
                sys.exit(1)
        else:
            logger.warning(
                "Received administrative record of unknown type from '%s'!",
                msg.eid
            )

        counter += 1
        if max_count and counter >= max_count:
            logger.info("Expected amount of bundles received, terminating.")
            return


def main():
    parser = argparse.ArgumentParser(
        description="register an agent with uD3TN and wait for bundles",
    )

    add_common_parser_arguments(parser)

    parser.add_argument(
        "-c", "--count",
        type=int,
        default=None,
        help="amount of bundles to be received before terminating",
    )
    parser.add_argument(
        "-o", "--output",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
        help="file to write the received bundle contents",
    )
    parser.add_argument(
        "--verify-pl",
        default=None,
        help="verify that the payload is equal to the provided string",
    )
    parser.add_argument(
        "--newline",
        action="store_true",
        help="print a line feed character after every received bundle payload",
    )

    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    try:
        if args.tcp:
            with AAPTCPClient(address=args.tcp) as aap_client:
                aap_client.register(args.agentid)
                run_aap_recv(
                    aap_client,
                    args.count,
                    args.output,
                    args.verify_pl,
                    args.newline,
                )
        else:
            with AAPUnixClient(address=args.socket) as aap_client:
                aap_client.register(args.agentid)
                run_aap_recv(
                    aap_client,
                    args.count,
                    args.output,
                    args.verify_pl,
                    args.newline,
                )
    finally:
        args.output.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging

from ud3tn_utils.aap2 import (
    AAP2TCPClient,
    AAP2UnixClient,
    AuthType,
    BundleADU,
    BundleADUFlags,
    ResponseStatus,
)
from ud3tn_utils.config import (
    JSONConfigMessage as ConfigMessage,
    make_contact,
    RouterCommand,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    argparse_non_empty_str,
    get_config_eid,
    get_secret_from_args,
    initialize_logger,
)
from pyd3tn.eid import validate_eid


logger = logging.getLogger(__name__)


def main():
    import sys

    parser = argparse.ArgumentParser(
        description="configure a contact via uD3TN's AAP 2.0 interface",
    )
    add_common_parser_arguments(parser)
    parser.add_argument(
        "--dest_eid",
        type=argparse_non_empty_str,
        default=None,
        help="the EID of the node to which the configuration belongs",
    )
    parser.add_argument(
        "eid",
        type=argparse_non_empty_str,
        nargs="?",
        default=None,
        help="the EID of the node to which the contact exists",
    )
    parser.add_argument(
        "cla_address",
        type=argparse_non_empty_str,
        nargs="?",
        default=None,
        help="the CLA address of the node",
    )
    parser.add_argument(
        "-s", "--schedule",
        nargs=3,
        type=int,
        metavar=("start_offset", "duration", "bitrate"),
        action="append",
        default=[],
        help="schedule a contact relative to the current time",
    )
    parser.add_argument(
        "-r", "--reaches",
        type=argparse_non_empty_str,
        action="append",
        default=[],
        help="specify an EID reachable via the node",
    )
    parser.add_argument(
        "-d", "--delete",
        action="store_true",
        help="perform a DELETE instead of an ADD operation",
    )
    parser.add_argument(
        "-q", "--query",
        action="store_true",
        help="perform a QUERY instead of an ADD operation",
    )
    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    if not args.query and not args.delete:
        if not args.schedule:
            logger.fatal("At least one -s/--schedule argument must be given!")
            sys.exit(1)
        if not args.cla_address:
            logger.fatal("A CLA address must be given!")
            sys.exit(1)
    else:
        if args.query and args.delete:
            logger.fatal("--query and --delete are mutually exclusive!")
            sys.exit(1)
    if not args.query and not args.eid:
        logger.fatal("An EID must be given!")
        sys.exit(1)
    for eid in [args.eid, args.dest_eid]:
        if eid:
            validate_eid(eid)
    for eid in args.reaches:
        validate_eid(eid)

    msg = bytes(ConfigMessage(
        args.eid,
        args.cla_address,
        contacts=[
            make_contact(*contact)
            for contact in args.schedule
        ],
        reachable_eids=args.reaches,
        type=(
            RouterCommand.DELETE if args.delete
            else (
                RouterCommand.QUERY if args.query
                else RouterCommand.ADD
            )
        ),
    ))

    logger.debug("> %s", msg)

    if args.tcp:
        aap2_client = AAP2TCPClient(address=args.tcp)
    else:
        aap2_client = AAP2UnixClient(address=args.socket)
    with aap2_client:
        secret = aap2_client.configure(
            args.agentid,
            subscribe=False,
            secret=get_secret_from_args(args),
            auth_type=AuthType.AUTH_TYPE_BUNDLE_DISPATCH,
        )
        logger.info("Assigned agent secret: '%s'", secret)
        dest_eid = args.dest_eid or aap2_client.node_eid
        aap2_client.send_adu(
            BundleADU(
                dst_eid=get_config_eid(dest_eid),
                payload_length=len(msg),
                adu_flags=[BundleADUFlags.BUNDLE_ADU_WITH_BDM_AUTH],
            ),
            msg,
        )
        response = aap2_client.receive_response()
        if response.response_status != ResponseStatus.RESPONSE_STATUS_SUCCESS:
            logger.warning(
                "µD3TN responded with unexpected status: %s (%d)",
                aap2_client.response_status_name(response.response_status),
                response.response_status,
            )
            sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

"""Script to send a bundle and receive a specified number of bundles."""

import argparse
import logging
import io
import sys

import cbor2

from pyd3tn.bundle7 import Bundle
from ud3tn_utils.aap import (
    AAPClient,
    AAPTCPClient,
    AAPUnixClient,
    AAPMessageType,
)
from ud3tn_utils.aap.bin.helpers import (
    add_common_parser_arguments,
    initialize_logger,
)


logger = logging.getLogger(__name__)


def _send_and_recv(aap_client: AAPClient, dest_eid: str, payload: bytes,
                   count: int, output: io.BufferedIOBase,
                   logger: logging.Logger):
    logger.info("Sending payload of length %d to %s", len(payload), dest_eid)
    aap_client.send_bundle(dest_eid, payload, False)
    for i in range(count):
        logger.info("Waiting for reception of bundle #%d of %d", i + 1, count)
        while True:
            try:
                msg = aap_client.receive()
            except KeyboardInterrupt:
                logger.info("Terminated by keyboard interrupt.")
                sys.exit(130)  # exit status for SIGINT
            if not msg:
                logger.warning("µD3TN has closed the connection.")
                sys.exit(1)
            enc = False
            if msg.msg_type == AAPMessageType.RECVBUNDLE:
                payload = msg.payload
            elif msg.msg_type == AAPMessageType.RECVBIBE:
                payload = cbor2.loads(msg.payload)
                bundle = Bundle.parse(payload[2])
                payload = bundle.payload_block.data
                enc = True
            else:
                continue
            enc_str = " encapsulated" if enc else ""
            logger.info(
                "Received%s bundle from '%s' with payload of len = %d",
                enc_str,
                msg.eid,
                len(payload),
            )
            output.write(payload)
            output.flush()
            break
    logger.info("Expected number of bundles received, terminating")


def _parse_args():
    parser = argparse.ArgumentParser(
        description="send a bundle via uD3TN and wait for responses",
    )
    add_common_parser_arguments(parser)
    parser.add_argument(
        "DESTINATION",
        help="the destination EID of the created bundle",
    )
    parser.add_argument(
        "PAYLOAD",
        type=argparse.FileType("rb"),
        default=sys.stdin.buffer,
        nargs="?",
        help="the payload of the created bundle, (default: read from STDIN)",
    )
    parser.add_argument(
        "-c", "--count",
        type=int,
        default=1,
        help="count of bundles expected to be received (default: 1)",
    )
    parser.add_argument(
        "-o", "--output",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
        help="file to write the received bundle contents",
    )
    return parser.parse_args()


def main():
    args = _parse_args()
    global logger
    logger = initialize_logger(args.verbosity)

    payload = args.PAYLOAD.read()
    args.PAYLOAD.close()

    try:
        if args.tcp:
            aap_client = AAPTCPClient(address=args.tcp)
        else:
            aap_client = AAPUnixClient(address=args.socket)
        aap_client.connect()
        try:
            aap_client.register(args.agentid)
            _send_and_recv(
                aap_client,
                args.DESTINATION,
                payload,
                args.count,
                args.output,
                logger,
            )
        finally:
            aap_client.disconnect()
    finally:
        args.output.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import logging
import sys

DEFAULT_CONFIG_AGENT_ID_DTN = "config"
DEFAULT_CONFIG_AGENT_ID_IPN = "9000"


def initialize_logger(verbosity: int):
    log_level = {
        0: logging.WARN,
        1: logging.INFO,
    }.get(verbosity, logging.DEBUG)
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s %(levelname)s: %(message)s",
    )
    return logging.getLogger(sys.argv[0])


def add_common_parser_arguments(parser):
    add_socket_group_parser_arguments(parser)
    add_agentid_parser_argument(parser)
    add_verbosity_parser_argument(parser)


def add_verbosity_parser_argument(parser):
    parser.add_argument(
        "-v", "--verbosity",
        action="count",
        default=0,
        help="increase output verbosity"
    )


def add_agentid_parser_argument(parser):
    parser.add_argument(
        "-a", "--agentid",
        default=None,
        help="the agent id to register with uD3TN (default: random UUID)",
    )


def add_socket_group_parser_arguments(parser):
    socket_group = parser.add_mutually_exclusive_group()
    socket_group.add_argument(
        "--socket",
        metavar="PATH",
        default="ud3tn.socket",
        type=str,
        help=(
            "AAP UNIX domain socket to connect to "
            "(default: ud3tn.socket)"
        ),
    )
    socket_group.add_argument(
        "--tcp",
        nargs=2,
        metavar=("IP", "PORT"),
        type=str,
        help="AAP TCP socket to connect to",
    )


def get_node_eid_prefix(eid):
    if eid[0:6] == "dtn://":
        return "dtn://" + eid.split("/")[2] + "/"
    elif eid[0:4] == "ipn:":
        return eid.split(".")[0] + "."
    else:
        raise ValueError("Cannot determine the node prefix for the given EID.")


def get_config_eid(eid):
    return get_node_eid_prefix(eid) + (
        DEFAULT_CONFIG_AGENT_ID_DTN
        if eid[0] == "d"  # get_node_eid_prefix already checks everything else
        else DEFAULT_CONFIG_AGENT_ID_IPN
    )



# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0

from .generated.storage_agent_pb2 import (
    BundleMetadataFilter,
    CompoundBundleId,
    StorageCall,
    StorageOperation,
)

__all__ = [
    "BundleMetadataFilter",
    "CompoundBundleId",
    "StorageCall",
    "StorageOperation",
]



# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8
from .aap2_client import (
    AAP2AsyncClient,
    AAP2AsyncTCPClient,
    AAP2AsyncUnixClient,
    AAP2SyncClient,
    AAP2TCPClient,
    AAP2UnixClient,
    AAP2Error,
    AAP2CommunicationError,
    AAP2ServerDisconnected,
    AAP2OperationFailed,
    AAP2UnexpectedMessage,
)
from .generated.aap2_pb2 import (
    AAPMessage,
    AAPResponse,
    AuthType,
    Bundle,
    BundleADU,
    BundleADUFlags,
    BundleDispatchInfo,
    ConnectionConfig,
    DispatchReason,
    DispatchEvent,
    DispatchResult,
    Keepalive,
    Link,
    LinkFlags,
    LinkStatus,
    ResponseStatus,
    Welcome,
)

__all__ = [
    "AAP2AsyncClient",
    "AAP2AsyncTCPClient",
    "AAP2AsyncUnixClient",
    "AAP2SyncClient",
    "AAP2TCPClient",
    "AAP2UnixClient",
    "AAP2Error",
    "AAP2CommunicationError",
    "AAP2ServerDisconnected",
    "AAP2OperationFailed",
    "AAP2UnexpectedMessage",

    "AAPMessage",
    "AAPResponse",
    "AuthType",
    "Bundle",
    "BundleADU",
    "BundleADUFlags",
    "BundleDispatchInfo",
    "ConnectionConfig",
    "DispatchReason",
    "DispatchEvent",
    "DispatchResult",
    "Keepalive",
    "Link",
    "LinkFlags",
    "LinkStatus",
    "ResponseStatus",
    "Welcome",
]



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause OR Apache-2.0
# encoding: utf-8

import argparse
import logging
import sys

from ud3tn_utils.aap2 import (
    AAP2TCPClient,
    AAP2UnixClient,
    AAPMessage,
    AuthType,
    Link,
    LinkFlags,
    LinkStatus,
    ResponseStatus,
)
from ud3tn_utils.aap2.bin.helpers import (
    add_common_parser_arguments,
    argparse_non_empty_str,
    get_secret_from_args,
    initialize_logger,
)
from pyd3tn.eid import validate_eid


logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description="configure a FIB entry via AAP 2.0",
    )
    add_common_parser_arguments(parser)
    parser.add_argument(
        "NODE_ID",
        type=argparse_non_empty_str,
        help="the next-hop node ID to be configured",
    )
    parser.add_argument(
        "CLA_ADDR",
        type=argparse_non_empty_str,
        help="the CLA address used to reach the next hop",
    )
    parser.add_argument(
        "-i", "--indirect",
        action="store_true",
        help="do not set the DIRECT flag (require BDM dispatch for the link)",
    )
    parser.add_argument(
        "-d", "--delete",
        action="store_true",
        help="if set, the entry will be deleted",
    )
    args = parser.parse_args()
    global logger
    logger = initialize_logger(args.verbosity)
    validate_eid(args.NODE_ID)

    if args.tcp:
        aap2_client = AAP2TCPClient(address=args.tcp)
    else:
        aap2_client = AAP2UnixClient(address=args.socket)
    with aap2_client:
        secret = aap2_client.configure(
            args.agentid,
            subscribe=False,
            secret=get_secret_from_args(args),
            auth_type=AuthType.AUTH_TYPE_FIB_CONTROL,
        )
        logger.info("Assigned agent secret: '%s'", secret)
        if args.delete:
            target_link_status = LinkStatus.LINK_STATUS_DOWN
        else:
            target_link_status = LinkStatus.LINK_STATUS_UP
        logger.info(
            "Sending link update for '%s' via '%s': %s",
            args.NODE_ID,
            args.CLA_ADDR,
            target_link_status,
        )
        aap2_client.send(AAPMessage(link=Link(
            status=target_link_status,
            flag=(
                LinkFlags.LINK_FLAG_NONE if args.indirect
                else LinkFlags.LINK_FLAG_DIRECT
            ),
            peer_node_id=args.NODE_ID,
            peer_cla_addr=args.CLA_ADDR,
        )))
        status = aap2_client.receive_response().response_status
        if status != ResponseStatus.RESPONSE_STATUS_SUCCESS:
            logger.warning(
                "µD3TN responded with unexpected status: %s (%d)",
                aap2_client.response_status_name(status),
                status,
            )
            sys.exit(1)


if __name__ == "__main__":
    main()
